'''
This program is intended to identify the V-gene and J-gene contributions to a given CDR3 Alpha sequence as well as identify any potential sequencing errors in the CDR3 and attempt to fix them
'''

import re, os, sys, csv
from Bio import SeqIO
from Bio.Seq import Seq

J_nt_path, J_aa_path = 'D:\\Old Computer\\Harsha\\Documents\\Immunology\\Pippa\\Beginning_of_J_Alpha_Genes.txt', 'D:\\Old Computer\\Harsha\\Documents\\Immunology\\Pippa\\TRAJ 3 aa before F.txt'
V_nt_path, V_aa_path = 'D:\\Old Computer\\Harsha\\Documents\\Immunology\\Kappler\\Harsha_454_Files\\B6_End_of_V_Alpha_Genes.txt', 'D:\\Old Computer\\Harsha\\Documents\\Immunology\\Kappler\\Harsha_454_Files\\B6_End_of_V_Alpha_Genes_AA.txt'

'''
file_path = 'D:\\Old Computer\\Harsha\\Documents\\Immunology\\Pippa\\'
file_input = 'modifying_test.txt' #THIS IS WHERE YOU HAVE TO ENTER THE INPUT FILE NAME BETWEEN THE APOSTROPHES (HAS TO BE A .csv FILE)
file_output = 'full-report.csv' #THIS IS WHERE YOU HAVE TO ENTER THE OUTPUT FILE NAME BETWEEN THE APOSTROPHES (HAS TO BE A .csv FILE)
'''
all_seqs = []
f = open('F:\\Dan14-405VJ.TXT', 'r')
for line in f:
    all_seqs.append(line)
f.close()
'''
f = open('F:\\DAN-12-205\\12205VJ2.TXT', 'r')
for line in f:
    if re.search('Vfam', line) or line == '\n':
        continue
    all_seqs.append(line)
f.close()
'''
g_code = { 'TTT': 'F', 'TCT': 'S', 'TAT': 'Y', 'TGT': 'C',
    'TTC': 'F', 'TCC': 'S', 'TAC': 'Y', 'TGC': 'C',
    'TTA': 'L', 'TCA': 'S', 'TAA': '*', 'TGA': '*',
    'TTG': 'L', 'TCG': 'S', 'TAG': '*', 'TGG': 'W',
    'CTT': 'L', 'CCT': 'P', 'CAT': 'H', 'CGT': 'R',
    'CTC': 'L', 'CCC': 'P', 'CAC': 'H', 'CGC': 'R',
    'CTA': 'L', 'CCA': 'P', 'CAA': 'Q', 'CGA': 'R',
    'CTG': 'L', 'CCG': 'P', 'CAG': 'Q', 'CGG': 'R',
    'ATT': 'I', 'ACT': 'T', 'AAT': 'N', 'AGT': 'S',
    'ATC': 'I', 'ACC': 'T', 'AAC': 'N', 'AGC': 'S',
    'ATA': 'I', 'ACA': 'T', 'AAA': 'K', 'AGA': 'R',
    'ATG': 'M', 'ACG': 'T', 'AAG': 'K', 'AGG': 'R',
    'GTT': 'V', 'GCT': 'A', 'GAT': 'D', 'GGT': 'G',
    'GTC': 'V', 'GCC': 'A', 'GAC': 'D', 'GGC': 'G',
    'GTA': 'V', 'GCA': 'A', 'GAA': 'E', 'GGA': 'G',
    'GTG': 'V', 'GCG': 'A', 'GAG': 'E', 'GGG': 'G'               
    }

global J_nt_dct, J_aa_dct, V_nt_dct, V_aa_dct
J_nt_dct, J_aa_dct, V_nt_dct, V_aa_dct = {}, {}, {}, {}

f = open(J_nt_path, 'r')
for item in list(SeqIO.parse(f, 'fasta')):
    J_nt_dct[item.id] = item.seq
f.close()
f = open(J_aa_path, 'r')
for item in list(SeqIO.parse(f, 'fasta')):
    J_aa_dct[item.id] = item.seq
f.close()

f = open(V_nt_path, 'r')
for item in list(SeqIO.parse(f, 'fasta')):
    V_nt_dct[item.id] = item.seq
f.close()
f = open(V_aa_path, 'r')
for item in list(SeqIO.parse(f, 'fasta')):
    V_aa_dct[item.id] = item.seq
f.close()

complementary = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}

#function identifies any palindromic nucleotides in the CDR3
def getPalindromes(seq, gene_seq, check):
    pal_NTs = ''
    if check:
        for i in range(len(gene_seq)):
            if complementary[gene_seq[-i-1]] != seq[i]:
                break
            pal_NTs += seq[i]
    else:
        for i in range(len(seq)):
            if complementary[gene_seq[-i-1]] != seq[i]:
                break
            pal_NTs += seq[i]

    return pal_NTs
    
#function identifies what the different regions (V, J, n-nucleotide, p-nucleotide) are in a sequence
def getRegions(V_gene, J_gene, seq):
    V_region, J_region, Vp, Jp, n_region = '', '', '', '', ''
    V_seq, J_seq = str(V_nt_dct[V_gene]), str(J_nt_dct[J_gene])

    for i in range(min(len(V_seq), len(seq))):
        if V_seq[i] != seq[i]:
            break
        V_region += V_seq[i]

    if len(seq)-len(V_region) < len(J_seq):
        for i in range(1, len(seq)-len(V_region) + 1):
            if J_seq[-i] != seq[-i]: #subtract 4 for seq because you want to remove TTT/TTC
                break
            J_region = J_seq[-i] + J_region
    else:
        for i in range(1, len(J_seq)+1):
            if J_seq[-i] != seq[-i]:
                break
            J_region = J_seq[-i] + J_region

    pnp = seq[len(V_region):len(seq)-len(J_region)]

    if len(V_region) == len(V_seq):
        Vp = getPalindromes(pnp, V_seq, len(pnp) > len(V_seq))
        pnp = pnp[len(Vp):]
    if len(J_region) == len(J_seq):
        Jp = getPalindromes(pnp[::-1], J_seq[::-1], len(pnp) > len(J_seq))[::-1]
        pnp = pnp[:len(pnp)-len(Jp)]
    n_region = pnp

    return[V_region, Vp, n_region, Jp, J_region]

def translate(seq):
    seq_trans = ''
    if len(seq)%3 == 0:
        for i in range(0, len(seq), 3):
            seq_trans += g_code[seq[i:i+3]]
    else:
        seq_trans = 'Sequence length not divisible by 3'

    return seq_trans

#function modifies a CDR3 if there are sequencing errors present in the sequence if the 3 nucleotides after the mismatch with the germline sequence continue to match the germline sequence
def fixCDR3(seq, gene, VJ):
    temp = seq
    if VJ == 'V':
        comp_seq = str(V_nt_dct[gene])
    else:
        comp_seq = str(J_nt_dct[gene])[::-1]

    
    for i in range(min(len(temp), len(comp_seq))-3):
        if comp_seq[i] != temp[i] and comp_seq[i+1:i+4] == temp[i+1:i+4]:
            temp = comp_seq[:i+4] + temp[i+4:]
            break
    return temp

def getIndel(seq, gene, VJ):
    mismatch, longest, ret_seqs = -1, [0, 0, 0, 0], ['', '', '', '']
    if VJ == 'V':
        check_seq = str(V_nt_dct[gene])
    else:
        check_seq = str(J_nt_dct[gene])[::-1]

    for i in range(min(len(check_seq), len(seq))-4):
        if check_seq[i] != seq[i]:
            mismatch = i
            break
    
    #Insertions in the CDR3
    if check_seq[mismatch:mismatch+3] == seq[mismatch+1:mismatch+4]:
        ret_seqs[0], temp = (seq[:mismatch] + seq[mismatch+1:]), 3
        for j in range(mismatch+3, min(len(check_seq), len(seq))-1):
            if check_seq[j] == seq[j+1]:
                temp += 1
            else:
                break
        longest[0] = temp
    if check_seq[mismatch:mismatch+3] == seq[mismatch+2:mismatch+5]:
        ret_seqs[1], temp = (seq[:mismatch] + seq[mismatch+2:]), 3
        for j in range(mismatch+3, min(len(check_seq), len(seq))-2):
            if check_seq[j] == seq[j+2]:
                temp += 1
            else:
                break
        longest[1] = temp

    #Deletions in the CDR3
    if (mismatch+4) <= len(check_seq) and check_seq[mismatch+1:mismatch+4] == seq[mismatch:mismatch+3]:
        ret_seqs[2], temp = (check_seq[:mismatch+4] + seq[mismatch+3:]), 4
        for j in range(mismatch+4, min(len(check_seq), len(seq))):
            if check_seq[j] == seq[j-1]:
                temp += 1
            else:
                break
        longest[2] = temp
    if (mismatch+5) <= len(check_seq) and check_seq[mismatch+2:mismatch+5] == seq[mismatch:mismatch+3]:
        ret_seqs[3], temp = (check_seq[:i+5] + seq[i+3:]), 5
        for j in range(mismatch+5, min(len(check_seq), len(seq))):
            if check_seq[j] == seq[j-2]:
                temp += 1
            else:
                break
        longest[3] = temp

    if max(longest) != 0:
        return ret_seqs[longest.index(max(longest))]
    else:
        return seq

header, other_seqs, count, valid_seqs = [], [], 0, {}
for line in all_seqs:
    temp = re.sub(' |\n', '', line).split(',')
    if re.search('Origin', line):
        header.append(temp[5:11] + ['V-Region', 'V-Palindromic-NTs', 'n-Region', 'J-Palindromic-NTs', 'J-Region', 'Frequency'])
        other_seqs.append(temp)
        continue
    if line == '\n' or temp[8] == '4-Unk' or temp[5]+temp[6] not in V_nt_dct or temp[7] not in J_nt_dct: #only look at seqs which are not unknown and have a V and J in the ref dictionaries
        other_seqs.append(temp)
        continue
    if temp[8] == '3-OOF':
        new_cdr3 = getIndel(temp[10][::-1], temp[7], 'J')[::-1]
        if temp[6] == 'Unk':
            other_seqs.append(temp[:9] + [new_cdr3] + temp[11:])
            continue
        else:
            if len(new_cdr3)%3 != 0:
                new_cdr3 = getIndel(new_cdr3, temp[5]+temp[6], 'V')
                if len(new_cdr3)%3 != 0:
                    other_seqs.append(temp[:9] + [new_cdr3] + temp[11:])
                    continue
                else:
                    cdr3_trans = translate(new_cdr3)
                    if cdr3_trans.find('*') != -1:
                        other_seqs.append(temp[:8] + [cdr3_trans, new_cdr3] + temp[11:])
                        continue
                    temp[10], temp[9] = new_cdr3, cdr3_trans
            else:
                cdr3_trans = translate(new_cdr3)
                if cdr3_trans.find('*') != -1:
                    other_seqs.append(temp[:8] + [cdr3_trans, new_cdr3] + temp[11:])
                    continue
                temp[10], temp[9] = new_cdr3, cdr3_trans


    else:
        temp[10] = fixCDR3(temp[10][::-1], temp[7], 'J')[::-1]
        temp[9] = translate(temp[10])
        #temp[9], temp[10] = re.sub(temp[9][-3:], str(J_aa_dct[str(temp[7])]), temp[9]), re.sub(temp[10][-9:], str(J_nt_dct[str(temp[7])])[-9:], temp[10])
        if temp[6] == 'Unk':
            temp[9] = 'C' + temp[9][1:]
            other_seqs.append(temp)
            continue
        else:
            temp[10] = fixCDR3(temp[10], temp[5]+temp[6], 'V')
            temp[9] = translate(temp[10])
            if temp[9].find('*') != -1:
                other_seqs.append(temp)
                continue
            #temp[9], temp[10] = re.sub(temp[9][:2], str(V_aa_dct[str(temp[5] + temp[6])])[:2], temp[9]), re.sub(temp[10][:6], str(V_nt_dct[str(temp[5] + temp[6])])[:6], temp[10])
            
    cdr3_info = getRegions(temp[5]+temp[6], temp[7], temp[10])
    all_info = temp[5] + ',' + temp[6] + ',' + temp[7] + ',' + temp[8] + ',' + temp[9] + ',' + temp[10] + ',' + cdr3_info[0] + ',' + cdr3_info[1] + ',' + cdr3_info[2] + ',' + cdr3_info[3] + ',' + cdr3_info[4]
    if all_info in valid_seqs:
        valid_seqs[all_info] += 1
    else:
        valid_seqs[all_info] = 1
    count += 1
    if count%1000 == 0:
        print 'Loop Parsing Counter: ' + str(count)

f.close()


f = open('F:\\Dan14_Valid Sequences.csv', 'wb')
writer = csv.writer(f)
writer.writerow(header[0])
for seq in valid_seqs:
    temp = seq.split(',')
    writer.writerow(temp + [valid_seqs[seq]])
f.close()

f = open('F:\\Dan14_Invalid_Sequences.csv', 'wb')
writer = csv.writer(f)
for line in other_seqs:
    writer.writerow(line)
f.close()


